package com.telelogic.tau.tor;

import java.util.LinkedList;

public class DispatchableClass extends Dispatchable
{
    // Attrmpts to send a signal through all existing ports
    // Returns true if the signal was send, false otherwise
    public boolean send(Event e)
    {
        for (Port p : m_listOutPorts)
        {
            if (p == null)
                continue;
            
            if (p.send(e))
                return true;
        }
        
        // Not received.
        return false;
    }

	// Attempts to dispatch an event on this dispatchable class.
    // Returns true if the event was consumed, false otherwise.	 
    // N.B. If the event is consumed the responsibility for the event is then at the receiver, while
	// if it is not consumed the responsibility for the event is then still at the sender.
	@Override
    public boolean receive(Event event)
    {
    	// If this instance is associated with an event queue, mark
    	// it as the receiver of the event, insert it into the queue
    	// and mark it as consumed
    	
    	if (m_eventQueue != null)
        {
            event.m_receiver = InstanceManager.instance().getIdFromExecutor(this);
            m_eventQueue.insert(event);
            return true; // Event consumed
        }
        return false; // Event not consumed
    }
    
    // Initializes this instance to make it ready for execution.
    // This typically involves creating the statemachine of the class 
    // and initialize it, and also create and initialize the architecture 
    // of the class.
    public void init()
    {
    	// Initialize the statemachine if found
    	StateMachine stateMachine = getClassifierBehavior();
    	if (stateMachine != null)
    		stateMachine.init();
    }

    // Initializes all dispatchable classes of the collection.
    public static <T extends DispatchableClass> void init(Iterable<T> collection)
    {
    	for (T t : collection)
		{
    		if (t != null)
    			t.init();
		}
    }
    
    public void finish()
    {
    	if (m_dispatcher != null) 
    	{
    		// Remove ourselves from the dispatcher
    		m_dispatcher.remove(this);
    		m_dispatcher = null;
    	}	
    }

    // Starts this instance and its statemachine, making it receptive to
    // events. The instance must be initialized before it is started.
    public void start()
    {
    	StateMachine stateMachine = getClassifierBehavior();
    	if (stateMachine != null)
    		stateMachine.start();
    }
    
    // Starts all dispatchable classes of the collection.
    public static <T extends DispatchableClass> void start(Iterable<T> collection)
    {
    	for (T t : collection)
		{
    		if (t != null)
    			t.start();
		}
    }

    // Executes the event in the statemachine of this class.
    // Returns the status to indicate how event was handled. 
    @Override
    public EventAction execute(Event e)
    {
    	m_currentEvent = e;
        EventAction ea = EventAction.NoMatch;
        StateMachine stateMachine = getClassifierBehavior();
        if (stateMachine != null)         
            ea = stateMachine.execute(e);
        
        m_currentEvent = null;
        return ea;
    }

    // Returns the (statemachine) behavior of this class.
    public StateMachine getClassifierBehavior()
    {
        return null;
    }

    // Returns this dispatchable class.
    public DispatchableClass getDispatchableClass() 
    {
        return this;
    }

    // Adds the given port as an out port of this dispatchable class.
    public void addOutPort(Port port)
    {
    	m_listOutPorts.add(port);
    }
    
    // List of outgoing ports defined for this dispatchable class.
    protected LinkedList<Port> m_listOutPorts = new LinkedList<Port>();
}
