package com.telelogic.tau.tor;

public abstract class Region
{
	public Region(State owner)
	{
		m_owner = owner;
		m_current = null;
		m_previous = null;
		m_finished = false;
	}

	// Executes an event by passing it to the current state. If there's no current
	// state, the event is passed to the state owning this region.
	// Returns EventAction.Consumed is the event was consumed (the responsibility for the event is then at the receiver),
	// or EventAction.NoMatch if no match was found (the responsibility for the event is then still at the sender).
	public Dispatchable.EventAction execute(Event e)
	{
		Dispatchable.EventAction result = Dispatchable.EventAction.NoMatch;

		// Handle event in current state if any.
		State state = m_current;

		// Otherwise use owner of this region.
		if (m_current == null) 
			state = m_owner;		

		if (state == null)
			return result;		

		StateMachine submachine = state.getStateMachine();
		if (submachine != null)
		{
			// Execute event in submachine
			result = submachine.execute(e);
		} 
		else 
		{
			// Loop over all regions and try to execute the event
			for (Region r : state.m_regions)
			{
				result = r.execute(e);
				if (result != Dispatchable.EventAction.NoMatch)
					// Event was consumed or deferred, evaluate transitions triggered only by guards.
					r.execute(null);
			}
		}

		// If event was not consumed by submachine nor regions, handle in current state
		if (result == Dispatchable.EventAction.NoMatch) 
		{
			// Event is about to be received!
			e.received(); 

			result = state.execute(e);
		}

		return result;
	}

	// Enters the region by entering its current state if any. The initial 
	// transition is used the first time or if there's no current state.
	void enter()
	{
		// Mark region as not finished.
		m_finished = false;

		// Enter region using currently active state. 
		// If there's no current state, use initial transition.
		if (m_current != null)
			m_current.enter();
		else 
			callInitialTransition();
	}

	// Enter the region using history, i.e. entering the previous state.
	// the default history transition is executed if there's no previous state.
	// The deepHistory flag specifies if shallow or deep history shall be used.
	// If the flag is true any composite states or submachines in the region 
	// will be recursively entered using history. Otherwise history only 
	// applies to this region.
	public void enterHistory() { enterHistory(false); }
	public void enterHistory(boolean deepHistory)
	{
		// Mark region as not finished.
		m_finished = false;

		// Enter previous state if any, otherwise use default history transition.
		if (m_previous != null) 
		{
			if (deepHistory)
				m_previous.enterHistory(deepHistory);
			else
				m_previous.enter();
		} 
		else 		    
			defaultHistoryTransition();		
	}

	// Leaves the region by leaving the current state.
	public void leave()
	{
		//	Leave the region by leaving the currently active state.
		if (m_current != null)
			m_current.leave();	
	}

	// Finishes this region by marking it as finished, and checks if all 
	// regions of the owning state are finished. In that case it calls
	// finish on the state as well.
	public void finish() { finish(null); }
	public void finish(ExitPoint ep)
	{
		if (ep != null && m_owner != null && ep.m_owner == this) 
		{
			StateMachine sm = m_owner.getStateMachine();
			if (sm != null)
				sm.leave(ep);
		} 
		else 
		{
			// Mark region as finished.
			m_finished = true;

			if (m_owner != null) 
			{
				// If all regions of the owning state are finished, finish the state.
				if (m_owner.isFinished())
					m_owner.finish();
			}
		}	
	}

	// Performs a call to 'initialTransition'. The call is synchronous.
	// Overload this function to make the call in an asynchronous way.
	public void callInitialTransition() 
	{
		initialTransition();
	}

	// The initial transition of the region. Executed the first time the region 
	// is entered. 
	// Implemented by subclasses.
	public void initialTransition() {}

	// The default history transition of the region. Executed if the region
	// is entered using history, but there is no previous state.
	// Implemented by concrete subclasses.
	public void defaultHistoryTransition() {}

	// Used to execute transitions from exit points of owned submachines. When a submachine
	// is left this operation is called passing the exit point as a parameter and any matching
	// transition is taken.
    // Implemented by concrete subclasses.
	public void exitPointTransitions(ExitPoint ep) {}

	// Used to execute transitions from entry points of the region.
    // Implemented by concrete subclasses.
	public void entryPointTransitions(EntryPoint ep) {}

	// A general Region doesn't know to which statemachine it belongs.
	// Overloaded by subclasses.
	public StateMachine getStateMachine()
	{
		return null;
	}
	
	//////////////////////////////////////////////////////////////////////////////////////////////////
	
	// The owning state. Events not handled by the region are passed on to the
	// owning state.
	State m_owner;

	// The currently active state of the region. Events received by the region 
	// are passed on to this state. Is set when entering a state in the region 
	// and cleared when leaving a state.
	State m_current;

	// The previously active state of the region. Used when entering the region 
	// using history. Is set when changing the current state, e.g. when a state 
	// is entered or left.
	State m_previous;

	// Flag stating if the regions is finished or not.
	boolean m_finished;	
}
