package com.telelogic.tau.tor;

import java.util.ArrayList;

public abstract class State
{
	public State(Region owner)
	{
		m_owner = owner;
	}
	
	// Execute an event in this state. Implements transitions in response to events
	// and guards.
	// Returns EventAction.Consumed is the event was consumed (the responsibility for the event is then at the receiver),
	// or EventAction.NoMatch if no match was found (the responsibility for the event is then still at the sender).
	// Must be implemented by concrete subclasses.
	public abstract Dispatchable.EventAction execute(Event e);

	// Initialize the state. Default implementation does nothing.
	// A derivate representing a composite state should override
	// this function and initialize the statemachine of the composite state.
	public void init() {}

	// Returns the statemachine of a composite state. If the state is not
	// a composite state, null is returned.
	public StateMachine getStateMachine() 
	{
		return null; 
	}

	// Enter the state, optionally using the supplied entry point.
	// If the state has a submachine, the submachine is entered, otherwise
	// all top level regions of the state are entered.
	public void enter() { enter(null); }
	public void enter(EntryPoint ep)
	{
		// Update the previous and currently active states of the owning region.
		// Note: Done before entering the submachine / owned regions
		// to get correct semantics.
		m_owner.m_previous = m_owner.m_current;
		m_owner.m_current = this;

		// Enter the submachine if present, otherwise enter all owned regions.
		StateMachine subMachine = getStateMachine();
		if (subMachine != null) 
			subMachine.enter(ep);		
		else 
		{
			for (Region r : m_regions)			
				r.enter();			
		}

		// Evaluate transitions triggered only by guards.
		execute(null);
	}

	// Same as enter, but in addition uses the source state of the transition
	// to calculate which regions/states that needs to be left/entered to
	// enter this state. Every region/state from the source state to the
	// least common ancestor of the source state and this state should be left,
	// and every state from the ancestor to this state should be entered.
	public void enterFrom(State s, EntryPoint ep)
	{
		assert false; // NOT IMPLEMENTED!
	}

	// Enter the state using history, i.e. enter all top level regions using history.
	public void enterHistory() { enterHistory(false); }
	public void enterHistory(boolean deepHistory)
	{
		// Set current and previous states of the owning region.
		// Note: Done before entering the submachine / owned regions
		// to get correct semantics.
		m_owner.m_previous = m_owner.m_current;
		m_owner.m_current = this;

		// Using history, enter the submachine if present, otherwise enter all owned regions.
		StateMachine subMachine = getStateMachine();
		if (subMachine != null) 
			subMachine.enterHistory(deepHistory);
		else 
		{
			for (Region r : m_regions)			
				r.enterHistory(deepHistory);	
		}
	}

	// Same as enterHistory, but use source state to calculate which regions/states
	// to enter/leave (same as for enterFrom).
	public void enterHistoryFrom(State s)
	{
		assert false; // NOT IMPLEMENTED!	
	}

	// Leave the state by leaving the submachine or all top level regions.
	public void leave()
	{
		// Leave the submachine if present otherwise leave all owned regions.
		StateMachine subMachine = getStateMachine();
		if (subMachine != null) 
			subMachine.leave();
		else 
		{
			for (Region r : m_regions)			
				r.leave();	
		}

		// Set current and previous states of the owning region.
		// Note: Done after leaving the submachine/ owned regions
		// to get correct semantics.
		m_owner.m_previous = m_owner.m_current;
		m_owner.m_current = null;	
	}

	// Leave the state for a specific target state, i.e. leave all states up to
	// the least common ancestor of this state and the target state.
	public void leaveFor(State s)
	{
		assert false; // NOT IMPLEMENTED!		
	}

	// Same as leaveFor, but use an exit point to calculate the least common
	// ancestor.
	public void leaveFor(ExitPoint ep)
	{
		assert false; // NOT IMPLEMENTED!		
	}

	// Finishes this state by generating a CompletedEvent and executing it.
	public void finish()
	{
		//	Create a completion event.
		CompletedEvent e = new CompletedEvent();

		// Execute event synchronously.
		execute(e);
	}

	// Returns true if this state is finished, i.e. all owned regions are 
	// finished, false otherwise.
	public boolean isFinished()
	{
		// If the state has a submachine, return the status of it.
		StateMachine submachine = getStateMachine();
		if (submachine != null)
			return submachine.isFinished();

		// Otherwise loop over owned regions and return false
		// if any region is not finished.
		for (Region r : m_regions)
		{
			if (!r.m_finished)
				return false;
		}

		// All regions are finished, return true.
		return true;	
	}

	// Adds a region to this state.
	void addRegion(Region r)
	{
		m_regions.add(r);
	}
	
	////////////////////////////////////////////////////////////////////////////////////
	
	// The region owning this state.
	Region m_owner;

	// List of regions owned by this state.
	ArrayList<Region> m_regions = new ArrayList<Region>();
}
