package com.telelogic.tau.tor;

import java.util.ArrayList;

public abstract class StateMachine
{
	// The state of operation of the statemachine.
    enum StateOfOperation {
        Initial,	    // Created, but not yet started
        Started,	    
        Finished
    };
    StateOfOperation stateOfOperation;

    // Construction
    public StateMachine()
    {
    	stateOfOperation = StateOfOperation.Initial;
    }

    // Executes an event by looping over the top regions to see if they can consume the event.
    // The first matching region consumes the event, deletes it and returns Consumed.
    // If no match is found, NoMatch is returned and the event is not deleted.
    public Dispatchable.EventAction execute(Event e)
    {
    	if (e == null)
    		return Dispatchable.EventAction.Consumed;

    	Dispatchable.EventAction rv = Dispatchable.EventAction.NoMatch;

    	// Loop over non-finished top level regions and try to dispatch the event
    	for (TopRegion r : m_regions)
    	{
    		if (!(r.m_finished)) 
    		{
    			rv = r.execute(e);
    			if (rv != Dispatchable.EventAction.NoMatch )
    				break;    // Event was consumed by this region
    		}
    	}

    	// If event wasn't consumed by any region and this statemachine is not 
    	// a submachine, then consume it. If this statemachine is a submachine 
    	// the event can be handled by the parent and shouldn't be consumed.
    	if (( rv == Dispatchable.EventAction.NoMatch ) && ( !isSubStateMachine() ) )     	
    		rv = Dispatchable.EventAction.Consumed;    		     	

    	return rv;
    }

    // Initializes the statemachine by creating regions, states, entry/exit points, etc.
    // After calling init(), the statemachine is ready to be started using start().
    // This operation should be implemented by the concrete subclasses of StateMachine.
    public void init() { };

    // Starts the statemachine instance, executing the initial transition of all top level
    // regions. If the target of an initial transition is a composite state or a submachine,
    // their initial transitions are executed and so on.
    // After the call to start, the statemachine is running and is responding to events.
    // If the statemachine is already started when this function is called, nothing happens.
    public void start()
    {
    	if (stateOfOperation == StateOfOperation.Started)
    		return; // Already started.

    	stateOfOperation = StateOfOperation.Started;

    	// Enter the top level regions without passing any entry point.
    	enter(null);	
    }

    // Force the statemachine instance to finish. Removes the owning dispatchable
    // from its dispatcher. As a result of this the statemachine will not receive
    // any more events, and is considered to be finished.
    // Note: Finishing the statemachine does not imply leaving it.
    public void finish()
    {
    	if (stateOfOperation == StateOfOperation.Finished)
    		return; // Already finished.

    	stateOfOperation = StateOfOperation.Finished;	

    	if (isSubStateMachine())
    		return; 
    	
    	// This is not a sub statemachine so check if it is a classifier behaviour statemachine.
    	// If so we should finish the dispatchable class too.
    	
    	DispatchableClass dispatchableClass = getDispatchableClass();
    	if (dispatchableClass == null) 
    		return;

    	dispatchableClass.finish();
    }

    // Returns true if the statemachine is finished, false otherwise. The statemachine is
    // finished iff all top level regions are finished.
    public boolean isFinished()
    {
    	// Check if any region is not finished and in that case return false.
    	for (TopRegion r : m_regions)
    	{
    		if (!r.m_finished)
    			return false;
    	}

    	// Otherwise return true.
    	return true;	
    }

    // Returns true if this statemachine is a submachine, false otherwise.
    // The getOwningState function is used to decide if it is a submachine or not.
    public boolean isSubStateMachine()
    {
    	return (getOwningState() != null);
    }

    // Enters the statemachine by entering all top level regions. If a valid entry point
    // is supplied, the regions will be entered using this entry point.
    public void enter() { enter(null); }
    public void enter(EntryPoint ep)
    {
    	// Try to enter all top level regions using the supplied entry point
    	for (TopRegion r : m_regions)
    	{
    		// Enter region through entry point if ep is owned by region,
    		// otherwise enter region without entry point.
    		if (ep != null && (ep.m_owner == r)) 
    			ep.m_owner.entryPointTransitions(ep);
    		else
    			r.enter();    			
    	}
    }

    // Enter the statemachine using history, i.e. enter all top level regions using history.
    // By default shallow (not deep) history is entered.
    public void enterHistory() { enterHistory(false); }
    public void enterHistory(boolean deepHistory)
    {
    	// Enter all top level regions using history
    	for (TopRegion r : m_regions)
    	{
    		// Enter region using history
    		r.enterHistory(deepHistory);
    	}
    }

    // Leaves the statemachine by leaving all non-finished top level regions. If an
    // exit point is supplied, the statemachine will be exited using this exit point.
    // after leaving the statemachine, any matching transition in the owning statemachine 
    // is executed by calling exitPointTransitions supplying the exit point as a parameter.
    public void leave() { leave(null); }
    public void leave(ExitPoint ep)
    {
    	// Leave all non-finished top level regions.
    	for (TopRegion r : m_regions)
    	{
    		// Enter region using history
    		if (!r.m_finished)
    			r.leave();
    	}    	
    	
    	// If this is a submachine and an exit point was specified,
    	// search the owning state for matching transitions.
    	if (isSubStateMachine() && (ep != null)) 
    	{
    		Region owningRegion = getOwningState().m_owner;
    		if (owningRegion != null)
    			owningRegion.exitPointTransitions(ep);
    	}
    }

    // Adds a new top level region to the statemachine.
    public void addRegion(TopRegion tr)
    {
    	m_regions.add(tr);
    }

    // Returns the dispatchable class in which the statemachine is contained.
    public abstract DispatchableClass getDispatchableClass();

    // Returns the owning statemachine. Returns 0 if this is a top-level statemachine.
    public StateMachine getOwningStateMachine()
    {
    	// Recursively retrieve the owning region until we reach a 
    	// top level region.

    	State owner = getOwningState();
    	if (owner != null) 
    	{
    		Region region = owner.m_owner;
    		while (region != null && !(region instanceof TopRegion)) 
    		{
    			State tmpState = region.m_owner;
    			if (tmpState != null) 
    				region = tmpState.m_owner;
    			else 
    				region = null;    		
    		}
    		if (region != null) 
    		{
    			// Return the statemachine of the top level region
    			TopRegion topRegion = (TopRegion) region;
    			return topRegion.getStateMachine();
    		}
    	}
    	return null;
    }
    
    // Returns the owning state. Returns 0 if this is a top-level statemachine.    
    public State getOwningState() 
    {
        return null;
    }
    
    // Implemented by subclasses.
    public void initialTransition() {}
    
	// Used to execute transitions from exit points of owned submachines. When a submachine
	// is left this operation is called passing the exit point as a parameter and any matching
	// transition is taken.
	public void exitPointTransitions(ExitPoint ep) {}

	// Used to execute transitions from entry points of the statemachine.
	public void entryPointTransitions(EntryPoint ep) {}

    // Send the event to this state machine's dispatchable class.
    public boolean send(Event e)
    {
        DispatchableClass dc = getDispatchableClass();
        if (dc == null)
            return false;

        return dc.send(e);
    }

    
    /////////////////////////////////////////////////////////////////////////////
    
    // A list of all top level regions of the statemachine.
    ArrayList<TopRegion> m_regions = new ArrayList<TopRegion>();
}
