# Diagram generator agent producing a call graph, consisting of a class diagram
# that shows the 'context' operation O with dependencies to all operations called
# from the implementation of O.

proc CallGraphView { triggeredBy timing context server agentParameters } { 
  upvar 1 $agentParameters ap

  if {![u2::IsKindOf $context "Operation"]} {
    std::Output "Error: Expected model context is an Operation.\n"
    return
  }

  set model [u2::GetModel $context]
  if { $model == 0 } {
    std::Output "Error: This operation is not part of a model.\n"
    return
  }

  # Expect 3 agent parameters (2 dictated by the framework, and 1 custom one)
  if { [llength $ap] != 3 } {
    std::Output "Erroneous use of CallGraphView. 3 parameters expected.\n"
    return
  }

  # Make sure the context operation has an implementation. Otherwise we cannot generate a call graph.
  set impl [GetImplementation $context]

  if { $impl == 0 } {
    std::Output "Error: This operation has no implementation. Cannot generate a call graph!\n"
    return
  }

  # First agent parameter is the class diagram we should generate into.
  # If it is NULL we are responsible for creating it.
  set diagram [lindex $ap 0]

  # Second agent parameter is the list of additional entities synthesized.
  # This agent use this parameter for dependencies it synthesizes.
  set synthesizedEntities [lindex $ap 1]

  # Third parameter is added by this particular diagram generator: 
  # recursive : Boolean
  set recursive [lindex $ap 2]

  if { $diagram == 0 } {
    # No diagram supplied by the framework. Create a new top-level class diagram.
    set diagram [u2::Create $model "ClassDiagram"]
  }

  if { $diagram == 0 } {
    std::Output "Error: Failed to create class diagram in model.\n"
    return
  }

  # Set diagram name
  u2::SetValue $diagram "Name" "Call graph for [u2::GetValue $context "Name"]"
  
  # Set-up values for agent out parameters (required by diagram generator framework)
  set $ap [lset ap 0 $diagram] 

  # Clean-up from previous runs by deleting all dependencies in synthesizedEntities
  foreach dep $synthesizedEntities {
    u2::Delete $dep
  }

  # Run query on Implementation of the operation
  set query [u2::FindByName $model "CallGraph::GetCalledOperations"]
  if { $query == 0 } {
    std::Output "Error: Cannot find GetCalledOperations query!\n"
    return
  }

  set depth 0
  ComputeAndCreateDependenciesForOperation $model $query $context synthesizedEntities $recursive $depth

  # Set-up values for agent out parameters (required by diagram generator framework)
  # Note that in case just one entity was synthesized we must add an additional entry to
  # make sure Tcl will interpret the result as a list (as opposed to one single result entity).
  # If no entities at all were synthesized, we need not set-up the result.
  if { [llength $synthesizedEntities] > 0 } {
    set synthesizedEntities [concat $synthesizedEntities 0]
    set $ap [lset ap 1 $synthesizedEntities]] 
  }

  # Call class diagram populator (visualize all synthesized dependencies)	
  set agent [u2::FindByGuid $model "@TTDDiagramAgents@ClassDiagramPopulator"]
  if { $agent == 0 } {
    std::Output "Error: Cannot find ClassDiagramPopulator agent!\n"
    return
  }

  # Visualize the context operation, and all synthesized dependencies
  set toVisualize [concat $context $synthesizedEntities 0]

  set agentParams [list $toVisualize 1 0 0 0]
  u2::InvokeAgent $model $agent $diagram agentParams

  # Call refresh symbol utility to refresh all symbols in the diagram.
  # This will a.o.t. compute and set correct size of symbols.
  set agent [u2::FindByGuid $model "@TTDDiagramAgents@RefreshSymbol"]
  if { $agent == 0 } {
    std::Output "Error: Cannot find RefreshSymbol agent!\n"
    return
  }

  set agentParams [list {}]
  u2::InvokeAgent $model $agent $diagram agentParams

  # Call layout agent (we use Sugiyama layout)
  set agent [u2::FindByGuid $model "@TTDDiagramAgents@SugiyamaLayout"]
  if { $agent == 0 } {
    std::Output "Error: Cannot find SugiyamaLayout agent!\n"
    return
  }

  set agentParams [list {} {} 20 1]
  u2::InvokeAgent $model $agent $diagram agentParams

  # Call line router agent (makes all lines in the diagram look nice)
  set agent [u2::FindByGuid $model "@TTDDiagramAgents@LineRouter"]
  if { $agent == 0 } {
    std::Output "Error: Cannot find LineRouter agent!\n"
    return
  }

  set agentParams [list {} 0]
  u2::InvokeAgent $model $agent $diagram agentParams
}

# Helper: Run the 'query' on the implementation of the 'operation' and create a dependency in the
# 'model' for each operation in the query result. Add created dependencies to the 'synthesizedEntities'
# list. If 'recurse' is true, do this recursively for all operations in the query result. 
# 'depth' is the current depth of the call-stack.

proc ComputeAndCreateDependenciesForOperation { model query operation synthesizedEntities recurse depth } {
  upvar 1 $synthesizedEntities entities

  set impl [GetImplementation $operation]
  
  if { $impl == 0 } {
    return
  }

  set queryParams [lappend queryParams {} ]
  u2::InvokeAgent $model $query $impl queryParams

  # Create a dependency from the operation to each operation found by the query.
  foreach calledOperation [lindex $queryParams 0] {
    if { $calledOperation == 0 } {
      continue
    }   

    set recursionDetected 0

    if { $recurse != 0 } {
      # Check if the calledOperation is already on the call-stack. 

      for {set x 1} {$x <= $depth} {incr x} {
        upvar $x operation operationOnStack
        if { $operationOnStack == $calledOperation } {
          set recursionDetected 1
          break
        }
      }
    }

    set dep [u2::Create $model "Dependency"]
    if { $dep == 0 } {
      std::Output "Error: Failed to create dependency in model.\n"
      return
    }
    u2::SetEntity $dep "Supplier" $calledOperation
    u2::SetEntity $dep "NonOwningClient" $operation

    lappend entities $dep

    if { $recurse != 0 && $recursionDetected == 0} {
      incr depth
      ComputeAndCreateDependenciesForOperation $model $query $calledOperation entities $recurse $depth
      incr depth -1
    }
  }
}

# Helper: Extract the implementation of the 'operation', if any.

proc GetImplementation { operation } {
  set impl [u2::GetEntity $operation "InlineMethod"]
  if { $impl == 0 } {
    # No in-line impl. found. Check for a referenced one.
    set impls [u2::GetReferringEntities $operation "ImplementedSignature"]
    if { [llength $impls] != 0 } {
      # Multiple implementations found. Take the first one!
      set impl [lindex $impls 0]
    }
  }

  return $impl
}

# Diagram generator agent producing a call sequence view, consisting of a sequence diagram
# that shows the 'context' operation O and all operations called (directly or indirectly by the 
# implementation of O as lifelines. Each call sequence is shown using message lines and terminate
# when finished or when a recursive call is detected.

proc CallSequenceView { triggeredBy timing context server agentParameters } { 
  upvar 1 $agentParameters ap

  if {![u2::IsKindOf $context "Operation"]} {
    std::Output "Error: Expected model context is an Operation.\n"
    return
  }

  set model [u2::GetModel $context]
  if { $model == 0 } {
    std::Output "Error: This operation is not part of a model.\n"
    return
  }

  # Expect 2 agent parameters (dictated by the framework)
  if { [llength $ap] != 2 } {
    std::Output "Erroneous use of CallSequenceView. 2 parameters expected.\n"
    return
  }

  # Make sure the context operation has an implementation. Otherwise we cannot generate a call graph.
  set impl [GetImplementation $context]

  if { $impl == 0 } {
    std::Output "Error: This operation has no implementation. Cannot generate a call sequence view!\n"
    return
  }

  # First agent parameter is the sequence diagram we should generate into.
  # If it is NULL we are responsible for creating it.
  set diagram [lindex $ap 0]

  # Second agent parameter is the list of additional entities synthesized.
  # This agent doesn't use this parameter although it creates an initial top-level package
  # where the sequence diagram is created.
  set synthesizedEntities [lindex $ap 1]

  set diagramOwner 0

  if { $diagram != 0 } {
    # This is the case when the diagram gets regenerated.
    # We can't really keep the same diagram, so just delete the owner element and generate a new diagram

    set diagramOwner [u2::GetOwner $diagram]
    if { $diagramOwner != 0 } {
      set toBeDeleted $diagramOwner
      set diagramOwner [u2::GetOwner $toBeDeleted]
      u2::Delete $toBeDeleted
      set diagram 0
    }
  }

  if { $diagram == 0 } {
    # No diagram supplied by the framework. Create a sequence diagram (and a top-level package
    # to enclose it, if not already present).
    set package [u2::FindByName $model "'Call sequence data'"]
    if { $package == 0 } {
      set package [u2::Create $model "Package"]
    }
    if { $package == 0 } {
      std::Output "Error: Failed to create package in model.\n"
      return
    }

    if { $diagramOwner == 0 } {
      set diagramOwner $package
    }

    u2::SetValue $package "Name" "Call sequence data"

    set agent [u2::FindByGuid $model "@TTDDiagramAgents@CreateTraceSequenceDiagram"]
    if { $agent == 0} {
      std::Output "Error: Failed to find CreateTraceSequenceDiagram agent. Cannot create sequence diagram in model.\n"
      return
    }
    set p [list "Call sequence for [u2::GetValue $context "Name"]" 0]
    u2::InvokeAgent $model $agent $diagramOwner p
    set diagram [lindex $p 1]
  }

  if { $diagram == 0 } {
    std::Output "Error: Failed to create sequence diagram in model.\n"
    return
  }
  
  # Set-up values for agent out parameters (required by diagram generator framework)
  set $ap [lset ap 0 $diagram] 

  # TODO: Clean-up from previous runs by deleting everything in the owning Interaction except the diagram itself.

  # Run query on Implementation of the operation
  set query [u2::FindByName $model "CallGraph::GetCalledOperations"]
  if { $query == 0 } {
    std::Output "Error: Cannot find GetCalledOperations query!\n"
    return
  }
 
  set traceAgent [u2::FindByGuid $model "@TTDDiagramAgents@TraceEventToSequenceDiagram"]
  if { $traceAgent == 0} {
    std::Output "Error: Failed to find TraceEventToSequenceDiagram agent. Cannot populate sequence diagram.\n"
    return
  }

  set depth 0
  TraceCallsForOperation $model $query $traceAgent $context $diagram $depth

  # Finalize sequence diagram
  set agent [u2::FindByGuid $model "@TTDDiagramAgents@FinalizeTraceSequenceDiagram"]
  if { $agent == 0} {
    std::Output "Error: Failed to find FinalizeTraceSequenceDiagram agent\n"
    return
  }

  u2::InvokeAgent $model $agent $diagram
 
}

# Helper: Run the 'query' on the implementation of the 'operation' and trace a call in the 'diagram'
# for each operation in the query result. This is done using the 'traceAgent'.
# Then repeat this recursively for all operations in the query result. 
# 'depth' is the current depth of the call-stack.

proc TraceCallsForOperation { model query traceAgent operation diagram depth} {
  set impl [GetImplementation $operation]
  
  if { $impl == 0 } {
    return
  }

  set queryParams [lappend queryParams {} ]
  u2::InvokeAgent $model $query $impl queryParams

  # Trace a message from the operation to each operation found by the query.
  foreach calledOperation [lindex $queryParams 0] {    
    if { $calledOperation == 0 } {
      continue
    } 

    # Check if the calledOperation is already on the call-stack. 
    set recursionDetected 0

    for {set x 1} {$x <= $depth} {incr x} {
      upvar $x operation operationOnStack
      if { $operationOnStack == $calledOperation } {
        set recursionDetected 1
        break
      }
    }

    # In case we detected a recursion, mark the message line accordingly.
    if { $recursionDetected == 1 } {
      set eventInstance "eventInstance = 'recursion' (. .)"
    } else {
      set eventInstance ""
    }

    set p [list "EventSent (. 'sender' = \"[u2::GetValue $operation "Name"]\", $eventInstance, messageId = \"1\" .) " ]
    u2::InvokeAgent $model $traceAgent $diagram p
    set p [list "EventReceived (. 'receiver' = \"[u2::GetValue $calledOperation "Name"]\", $eventInstance, messageId = \"1\" .) " ]
    u2::InvokeAgent $model $traceAgent $diagram p

    if { $recursionDetected == 0 } {
      incr depth
      TraceCallsForOperation $model $query $traceAgent $calledOperation $diagram $depth
      incr depth -1
    }
  }
}
